from bluer_sbc.parts.db import db_of_parts

from bluer_ugv.designs.rangin.parts import dict_of_parts

docs = [
    {
        "path": "../docs/rangin/parts.md",
        "items": db_of_parts.as_images(
            dict_of_parts,
            reference="repo",
        ),
        "macros": {
            "parts:::": db_of_parts.as_list(
                dict_of_parts,
                reference="repo",
                log=False,
            ),
        },
    },
]



from bluer_ugv.README.swallow.digital.design.arzhang4 import docs as arzhang4
from bluer_ugv.README.swallow.digital.design import (
    ethernet,
    joystick,
    mechanical,
    obsolete,
    operation,
    parts,
    terraform,
    testing,
    ultrasonic_sensor,
)


docs = (
    [
        {
            "path": "../docs/swallow/digital/design",
        },
        {
            "path": "../docs/swallow/digital/design/rpi-pinout.md",
        },
    ]
    + arzhang4.docs
    + ethernet.docs
    + joystick.docs
    + mechanical.docs
    + obsolete.docs
    + operation.docs
    + parts.docs
    + terraform.docs
    + testing.docs
    + ultrasonic_sensor.docs
)



import threading
from typing import Union, Dict
import time


from bluer_options import string

from bluer_ugv import env
from bluer_ugv.swallow.session.classical.leds import ClassicalLeds
from bluer_ugv.swallow.session.classical.ethernet.classes import ClassicalEthernet
from bluer_ugv.swallow.session.classical.setpoint.steering import (
    generate_left_and_right,
)
from bluer_ugv.logger import logger


class ClassicalSetPoint:
    def __init__(
        self,
        ethernet: ClassicalEthernet,
        leds: ClassicalLeds,
    ):
        self.speed: int = 0
        self.max_speed: int = 100
        self.steering: int = 0

        self.steering_expiry: float = time.time()

        self.started: bool = False

        self.ethernet = ethernet
        self.leds = leds

        self._lock = threading.Lock()

        logger.info(f"{self.__class__.__name__} created.")

    def get(
        self,
        what: str = "all",
    ) -> Union[int, bool, Dict[str, Union[int, bool]]]:
        with self._lock:
            self.speed = max(
                -self.max_speed,
                min(
                    self.max_speed,
                    self.speed,
                ),
            )

            if what == "all":
                return {
                    "speed": self.speed,
                    "max_speed": self.max_speed,
                    "started": self.started,
                    "steering": self.steering,
                }

            if what == "left":
                return generate_left_and_right(self.speed, self.steering)[0]

            if what == "right":
                return generate_left_and_right(self.speed, self.steering)[1]

            if what == "max_speed":
                return self.max_speed

            if what == "speed":
                return self.speed

            if what == "started":
                return self.started

            if what == "steering":
                return self.steering

            logger.error(f"{self.__class__.__name__}.get: {what} not found.")
            return 0

    def check_steering_expiry(
        self,
        log: bool = True,
    ) -> bool:
        with self._lock:
            if self.steering == 0:
                return False

            if self.steering_expiry > time.time():
                if log:
                    logger.info(
                        "setpoint will expire in {}.".format(
                            string.pretty_duration(
                                self.steering_expiry - time.time(),
                                largest=True,
                                short=True,
                                include_ms=True,
                            )
                        )
                    )
                return False

        if log:
            logger.info("setpoint expired.")

        self.put(
            what="steering",
            value=0,
        )

        return True

    def put(
        self,
        value: Union[int, bool, Dict[str, Union[int, bool]]],
        what: str = "all",
        log: bool = True,
        steering_expires_in: float = env.BLUER_UGV_SWALLOW_STEERING_EXPIRY,
    ):
        with self._lock:
            if what == "all":
                self.speed = min(100, max(-100, int(value["speed"])))
                self.max_speed = value.get("max_speed", 100)
                self.started = bool(value["started"])
                self.steering = min(100, max(-100, int(value["steering"])))
                self.steering_expiry = time.time() + steering_expires_in
                return

            if what == "max_speed":
                self.max_speed = min(100, max(-100, int(value)))
                if log:
                    logger.info(
                        "{}.put: max_speed={}".format(
                            self.__class__.__name__,
                            self.max_speed,
                        )
                    )
                return

            if what == "speed":
                self.speed = min(100, max(-100, int(value)))
                if log:
                    logger.info(
                        "{}.put: speed={}".format(
                            self.__class__.__name__,
                            self.speed,
                        )
                    )
                return

            if what == "started":
                self.started = bool(value)
                if log:
                    logger.info(
                        "{}.put: {}".format(
                            self.__class__.__name__,
                            "started" if value else "stopped",
                        )
                    )
                return

            if what == "steering":
                self.steering = min(100, max(-100, int(value)))
                self.steering_expiry = time.time() + steering_expires_in
                if log:
                    logger.info(
                        "{}.put: steering={}, expires in {}".format(
                            self.__class__.__name__,
                            self.steering,
                            string.pretty_duration(
                                steering_expires_in,
                                largest=True,
                                short=True,
                                include_ms=True,
                            ),
                        )
                    )
                return

            logger.error(f"{self.__class__.__name__}.put: {what} not found.")

    def start(self):
        self.put(
            {
                "speed": 0,
                "started": True,
                "steering": 0,
            }
        )

        logger.info(f"{self.__class__.__name__}.start")

    def stop(
        self,
        hard: bool = False,
    ):
        instructions = {
            "speed": 0,
            "started": False,
            "steering": 0,
        }
        if hard:
            instructions["max_speed"] = 0
        self.put(instructions)

        logger.info(
            "{}.stop({})".format(
                self.__class__.__name__,
                "hard" if hard else "",
            )
        )

        self.leds.set("red", False)
        self.leds.set("yellow", False)

    def update(self) -> bool:
        with self._lock:
            if self.started:
                self.leds.flash("red")

        return True



from bluer_ai.env import BLUER_AI_SBC_PASSWORD

from bluer_ugv.README.swallow.digital import algo, design

docs = (
    [
        {
            "path": "../docs/swallow/digital",
        },
        {
            "path": "../docs/swallow/digital/manual.md",
            "macros": {
                "password:::": BLUER_AI_SBC_PASSWORD,
            },
        },
    ]
    + design.docs
    + algo.docs
)



from bluer_objects.README.items import ImageItems

from bluer_ugv.README.swallow.consts import swallow_electrical_designs

docs = [
    {
        "path": "../docs/swallow/digital/design/steering-over-current-detection.md",
        "items": ImageItems(
            {
                f"{swallow_electrical_designs}/steering-over-current.png": f"{swallow_electrical_designs}/steering-over-current.svg",
            }
        ),
    },
]



from typing import Dict

Controls: Dict[str, Dict] = {
    "axes": {
        3: ("speed", -1),
        2: ("steering", -1),
    },
    "buttons": {
        9: "ultrasonic off",
        7: "ultrasonic on",
        8: "debug off",
        6: "debug on",
        3: "mode = none",
        0: "mode = action",
        1: "mode = training",
        12: "special key",
        14: "stop",
    },
    "special-buttons": {
        0: "exit",
        1: "shutdown",
        3: "reboot",
        4: "update",
    },
}



from bluer_algo.socket.connection import DEV_HOST

from bluer_options.env import abcli_hostname
from bluer_sbc.session.functions import reply_to_bash

from bluer_ugv.swallow.session.classical.config.classes import ClassicalConfig
from bluer_ugv.swallow.session.classical.ethernet.classes import ClassicalEthernet
from bluer_ugv.swallow.session.classical.ethernet.command import EthernetCommand
from bluer_ugv.swallow.session.classical.leds import ClassicalLeds
from bluer_ugv.swallow.session.classical.setpoint.classes import ClassicalSetPoint
from bluer_ugv.logger import logger


class ClassicalController:
    def __init__(
        self,
        config: ClassicalConfig,
        ethernet: ClassicalEthernet,
        leds: ClassicalLeds,
        setpoint: ClassicalSetPoint,
    ):
        self.config = config
        self.ethernet = ethernet
        self.leds = leds
        self.setpoint = setpoint

        self.special_key: bool = False

        logger.info(f"created {self.__class__.__name__}")

    def reply_to_bash(
        self,
        event: str,
    ) -> bool:
        logger.info(f'*"{event}" is selected.')

        if self.ethernet.enabled:
            self.ethernet.client.send(
                EthernetCommand(
                    action="keyboard",
                    data={
                        "sender": abcli_hostname,
                        "event": event,
                    },
                ),
                drain=True,
            )

        return reply_to_bash(event)

    def set_debug(self, value: bool):
        self.special_key = False

        self.config.set("debug_mode", value)

        logger.info(
            f'debug enabled, run "@swallow debug" on {DEV_HOST}.'
            if value
            else logger.info("debug disabled")
        )

    def set_special_key(self):
        self.special_key = True

        logger.info("🛠️ special key enabled.")

    def set_ultrasonic(self, value):
        self.special_key = False

        self.config.set("ultrasound_enabled", value)

    def update(self) -> bool:
        if self.special_key:
            self.leds.flash_all()

        return True



from bluer_objects.README.items import ImageItems

from bluer_ugv.README.swallow.consts import swallow_assets2
from bluer_ugv.swallow.session.classical.controller.keyboard.keys import ControlKeys


docs = [
    {
        "path": "../docs/swallow/digital/design/operation.md",
        "cols": 2,
        "items": ImageItems(
            {
                f"{swallow_assets2}/20251019_121811.jpg": "",
                f"{swallow_assets2}/20251019_121842.jpg": "",
            }
        ),
        "macros": {
            "keys:::": ControlKeys.as_table(),
        },
    },
]



from bluer_objects.README.consts import designs_url
from bluer_objects.README.items import ImageItems


docs = [
    {
        "path": f"../docs/swallow/digital/design/joystick{suffix}",
        "items": ImageItems(items),
    }
    for suffix, items in {
        "": {},
        "/validation.md": {},
        "/operation.md": {
            designs_url(
                "swallow/joystick/controls.png",
            ): designs_url(
                "swallow/joystick/controls.svg",
            )
        },
    }.items()
]



from bluer_objects.README.items import ImageItems

from bluer_ugv.README.swallow.consts import swallow_assets2

docs = [
    {
        "path": "../docs/swallow/digital/design/testing.md",
        "items": ImageItems(
            {
                f"{swallow_assets2}/20251201_172535~2_1.gif": "",
                f"{swallow_assets2}/20251203_112604.jpg": "",
            }
        ),
    },
]



from bluer_objects.README.items import ImageItems

from bluer_ugv.README.swallow.consts import swallow_assets2

docs = [
    {
        "path": "../docs/swallow/digital/design/terraform.md",
        "items": ImageItems(
            {
                f"{swallow_assets2}/20250611_100917.jpg": "",
                f"{swallow_assets2}/lab.png": "",
                f"{swallow_assets2}/lab2.png": "",
            }
        ),
    },
]



from typing import List

from bluer_options.terminal import show_usage, xtra


def help_install(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("dryrun", mono=mono)

    return show_usage(
        [
            "@ROS",
            "install",
            f"[{options}]",
        ],
        "install ROS.",
        mono=mono,
    )


def help_open(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("dryrun", mono=mono)

    return show_usage(
        [
            "@ROS",
            "open",
            f"[{options}]",
        ],
        "open ROS.",
        mono=mono,
    )


def help_start(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("~build,~cache,dryrun", mono=mono)

    return show_usage(
        [
            "@ROS",
            "start",
            f"[{options}]",
        ],
        "start ROS.",
        mono=mono,
    )


def help_stop(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("dryrun", mono=mono)

    return show_usage(
        [
            "@ROS",
            "stop",
            f"[{options}]",
        ],
        "stop ROS.",
        mono=mono,
    )


help_functions = {
    "install": help_install,
    "open": help_open,
    "start": help_start,
    "stop": help_stop,
}



import os

from bluer_options.help.functions import get_help
from bluer_objects import file, README

from bluer_ugv import NAME, VERSION, ICON, REPO_NAME
from bluer_ugv.help.functions import help_functions
from bluer_ugv.README.ugvs.comparison.build import build as build_comparison
from bluer_ugv.README.docs import docs


def build(args) -> bool:
    return (
        all(
            README.build(
                args=args,
                items=readme.get("items", []),
                path=os.path.join(file.path(__file__), readme["path"]),
                cols=readme.get("cols", 3),
                ICON=ICON,
                NAME=NAME,
                VERSION=VERSION,
                REPO_NAME=REPO_NAME,
                help_function=lambda tokens: get_help(
                    tokens,
                    help_functions,
                    mono=True,
                ),
                macros=readme.get("macros", {}),
            )
            for readme in docs
        )
        and build_comparison()
    )



from typing import Tuple
from enum import Enum, auto


class State(Enum):
    CLEAR = auto()
    WARNING = auto()
    DANGER = auto()

    @property
    def color_code(self) -> Tuple[int, int, int]:
        return (
            [0, 255, 0]
            if self == State.CLEAR
            else ([255, 255, 0] if self == State.WARNING else [255, 0, 0])
        )



import threading
import pygame

from bluer_options.logger import get_logger
from bluer_sbc.joystick import Joystick

from bluer_ugv.swallow.session.classical.config.classes import ClassicalConfig
from bluer_ugv.swallow.session.classical.controller.classes import ClassicalController
from bluer_ugv.swallow.session.classical.ethernet.classes import ClassicalEthernet
from bluer_ugv.swallow.session.classical.controller.joystick.controls import Controls
from bluer_ugv.swallow.session.classical.leds import ClassicalLeds
from bluer_ugv.swallow.session.classical.mode import OperationMode
from bluer_ugv.swallow.session.classical.setpoint.classes import ClassicalSetPoint
from bluer_ugv import ICON

logger = get_logger(f"{ICON}  🕹️")


class ClassicalJoystick(ClassicalController):
    def __init__(
        self,
        config: ClassicalConfig,
        ethernet: ClassicalEthernet,
        leds: ClassicalLeds,
        setpoint: ClassicalSetPoint,
    ):
        super().__init__(
            config=config,
            ethernet=ethernet,
            leds=leds,
            setpoint=setpoint,
        )

        self.joystick = Joystick()

        self.controls = Controls

        self.running: bool = False

        if not self.joystick.enabled:
            return

        self.running = True
        self.thread = threading.Thread(target=self.loop, daemon=True)
        self.thread.start()

    def handle_axis(
        self,
        axis: int,
        value: float,
    ):
        if axis in self.controls["axes"]:
            self.special_key = False

            axis_name, axis_direction = self.controls["axes"][axis]
            self.setpoint.put(
                what=axis_name,
                value=axis_direction * int(100 * value),
            )

    def handle_button(self, button: int):
        # bash keys
        if self.special_key:
            if button in self.controls["special-buttons"]:
                event = self.controls["special-buttons"][button]
                self.reply_to_bash(event)
                self.config.set("stop-requested", True)

        if button in self.controls["buttons"]:
            event = self.controls["buttons"][button]

            # ultrasonic
            if event == "ultrasonic off":
                self.set_ultrasonic(False)
            if event == "ultrasonic on":
                self.set_ultrasonic(True)

            # debug
            if event == "debug on":
                self.set_debug(True)
            if event == "debug off":
                self.set_debug(False)

            # mode
            mode = self.config.get("mode")
            updated_mode = mode
            if event == "mode = none":
                updated_mode = OperationMode.NONE

            if event == "mode = action":
                updated_mode = OperationMode.ACTION

            if event == "mode = training":
                updated_mode = OperationMode.TRAINING

            if mode != updated_mode:
                self.config.set("mode", updated_mode)
                self.special_key = False

            # stop
            if event == "stop":
                self.setpoint.stop()
                self.special_key = False

            # special key
            if event == "special key":
                self.set_special_key()

    def loop(self):
        logger.info(f"{self.__class__.__name__}.loop started.")

        while self.running:
            self.joystick.read_events()

            for event in self.joystick.buttons.values():
                if event.type == pygame.JOYBUTTONDOWN:
                    logger.info(
                        "{}: button {} pressed.".format(
                            self.__class__.__name__,
                            event.button,
                        )
                    )
                    self.handle_button(event.button)

            for event in self.joystick.axes.values():
                logger.info(
                    "{}: axis {} moved to {}.".format(
                        self.__class__.__name__,
                        event.axis,
                        event.value,
                    )
                )

                self.handle_axis(
                    axis=event.axis,
                    value=event.value,
                )

            if self.special_key:
                self.leds.flash_all()

            self.joystick.clock.tick(60)

    def stop(self):
        if not self.joystick.enabled:
            return

        self.running = False
        self.thread.join()

        logger.info(f"{self.__class__.__name__} stopped.")

        Joystick.cleanup()



import keyboard

from bluer_ugv.swallow.session.classical.config.classes import ClassicalConfig
from bluer_ugv.swallow.session.classical.controller.classes import ClassicalController
from bluer_ugv.swallow.session.classical.controller.keyboard.keys import ControlKeys
from bluer_ugv.swallow.session.classical.ethernet.classes import ClassicalEthernet
from bluer_ugv.swallow.session.classical.leds import ClassicalLeds
from bluer_ugv.swallow.session.classical.mode import OperationMode
from bluer_ugv.swallow.session.classical.setpoint.classes import ClassicalSetPoint
from bluer_ugv import env
from bluer_ugv.logger import logger


class ClassicalKeyboard(ClassicalController):
    def __init__(
        self,
        config: ClassicalConfig,
        ethernet: ClassicalEthernet,
        leds: ClassicalLeds,
        setpoint: ClassicalSetPoint,
    ):
        super().__init__(
            config=config,
            ethernet=ethernet,
            leds=leds,
            setpoint=setpoint,
        )

        self.keys = ControlKeys()

        self.last_key: str = ""

        self.is_used_for_steering: bool = False

        logger.info(f"created {self.__class__.__name__}")

    def update(self) -> bool:
        self.last_key = ""

        # bash keys
        if self.special_key:
            for key, event in self.keys.special_keys.items():
                if keyboard.is_pressed(key):
                    logger.info(f'*"{key}" is pressed.')
                    self.reply_to_bash(event)
                    return False

        # other keys
        for key, func in {
            self.keys.get("stop"): self.setpoint.stop,
            "x": self.setpoint.start,
            self.keys.get("speed backward"): lambda: self.setpoint.put(
                what="speed",
                value=self.setpoint.get(what="speed") - 10,
            ),
            self.keys.get("speed forward"): lambda: self.setpoint.put(
                what="speed",
                value=self.setpoint.get(what="speed") + 10,
            ),
        }.items():
            if keyboard.is_pressed(key):
                logger.info(f'"{key}" is pressed.')
                self.special_key = False
                func()

        # steering
        if keyboard.is_pressed(self.keys.get("steer left")):
            logger.info('"steer left" is pressed.')
            self.special_key = False
            self.last_key = "a"
            self.setpoint.put(
                what="steering",
                value=env.BLUER_UGV_SWALLOW_STEERING_SETPOINT,
            )

            self.is_used_for_steering = True
        elif keyboard.is_pressed(self.keys.get("steer right")):
            logger.info('"steer right" is pressed.')
            self.special_key = False
            self.last_key = "d"
            self.setpoint.put(
                what="steering",
                value=-env.BLUER_UGV_SWALLOW_STEERING_SETPOINT,
            )

            self.is_used_for_steering = True
        elif self.is_used_for_steering:
            if self.setpoint.check_steering_expiry():
                self.is_used_for_steering = False

        # debug mode
        if keyboard.is_pressed(self.keys.get("debug on")):
            self.set_debug(True)
        if keyboard.is_pressed(self.keys.get("debug off")):
            self.set_debug(False)

        # mode
        mode = self.config.get("mode")
        updated_mode = mode
        if keyboard.is_pressed(self.keys.get("mode = none")):
            updated_mode = OperationMode.NONE

        if keyboard.is_pressed(self.keys.get("mode = action")):
            updated_mode = OperationMode.ACTION

        if keyboard.is_pressed(self.keys.get("mode = training")):
            updated_mode = OperationMode.TRAINING

        if mode != updated_mode:
            self.config.set("mode", updated_mode)
            self.special_key = False

        # ultrasonic
        if keyboard.is_pressed(self.keys.get("ultrasonic off")):
            self.set_ultrasonic(False)
        if keyboard.is_pressed(self.keys.get("ultrasonic on")):
            self.set_ultrasonic(True)

        # special key
        if keyboard.is_pressed(self.keys.get("special key")) and not self.special_key:
            self.set_special_key()

        return super().update()
